/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Nickolay Ponomarev.
 *
 * Contributor(s):
 *  Nickolay Ponomarev <asqueella@gmail.com> (Original author)
 *  Malte Kraus <firefox@maltekraus.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const Ci = Components.interfaces, Cc = Components.classes, Cr = Components.results;

var globalObject = this;

function log(msg) {
  msg = "Adblock Filters:\t" + msg;
  var consoleService = Cc["@mozilla.org/consoleservice;1"]
                         .getService(Ci.nsIConsoleService);
  consoleService.logStringMessage(msg);
  //dump(msg + "\n");
  return msg;
}
function makeURI(uri, originCharset, baseURI) {
  var ioservice = Cc["@mozilla.org/network/io-service;1"]
                   .getService(Ci.nsIIOService);
  return ioservice.newURI(uri, originCharset || null, baseURI || null);
}
function loadScript(url, context) {
  if(url.indexOf(":/") == -1) url = "chrome://adblockfilters/content/" + url;
  if(!context && loadScript.caller && loadScript.caller != loadScriptOnce)
    context = loadScript.caller.__parent__;
  var ssLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"]
                   .getService(Ci.mozIJSSubScriptLoader);
  ssLoader.loadSubScript(url, context || null);
}
var loadScriptOnce = (function() {
  var loaded = {};
  return function _loadScriptOnce(url, context) {
    if(!context && _loadScriptOnce.caller)
      context = _loadScriptOnce.caller.__parent__;
    if(!(url in loaded) || loaded[url] != context) {
      loaded[url] = context;
      loadScript(url, context);
    }
  };
})();
loadScriptOnce("chrome://global/content/nsUserSettings.js");

function AdblockFilterService() {
  this.wrappedJSObject = this;
  this.prefs = nsPreferences;
  this._registerPrefObserver();
}
AdblockFilterService.prototype = {
  classID: Components.ID("{76c938ef-4597-2f95-4f1a-fad72f7ecb6c}"),
  contractID: "@adblockfilters.mozdev.org/service;1",
  classDescription: "Adblock Plus Filter Uploader service",

  wrappedJSObject: null,
  prefs: null,

  _closeObserverRegistered: false,
  _registerCloseObserver: function() {
    // does not work in Songbird which uses nsIAppStartup to kill the app
    if(this._closeObserverRegistered)
      return;
    this._closeObserverRegistered = true;

    Cc["@mozilla.org/observer-service;1"]
      .getService(Ci.nsIObserverService)
      .addObserver(this, "domwindowclosed", false);
  },
  _unregisterCloseObserver: function() {
    if(!this._closeObserverRegistered)
      return;
    this._closeObserverRegistered = false;

    Cc["@mozilla.org/observer-service;1"]
      .getService(Ci.nsIObserverService)
      .removeObserver(this, "domwindowclosed");
  },
  _registerPrefObserver: function() {
    var prefService = Cc["@mozilla.org/preferences-service;1"]
                        .getService(Ci.nsIPrefService);
    this._branch = prefService.getBranch("extensions.adblockfilters.");
    if(this._branch.getBoolPref("uploadOnExit"))
      this._registerCloseObserver();
    this._branch.QueryInterface(Ci.nsIPrefBranch2).addObserver("", this, false);
  },

  _unregisterPrefObserver: function() {
    if(this._branch) {
      this._branch.removeObserver("", this);
      this._branch = null;
    }
  },
  observe: function (aSubject, aTopic, aData) {
    if(aTopic == "nsPref:changed") {
      if(aData == "uploadOnExit") {
        if(aSubject.QueryInterface(Ci.nsIPrefBranch).getBoolPref("uploadOnExit"))
          this._registerCloseObserver();
        else
          this._unregisterCloseObserver();
      } else if(aData.indexOf("blacklist.") == 0) {
        this._blacklist = null;
      }
    } else if(aTopic == "domwindowclosed") {
      this._doLastUpload(aSubject);
    }
  },

  _doLastUpload: function(aWindow) {
    var windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"]
                           .getService(Ci.nsIWindowMediator);
    var win = windowMediator.getMostRecentWindow("");
    if(win) // there's at least one remaining window open
      return;

    this._unregisterCloseObserver();

    this.doUpload(aWindow);
  },

  /* utility functions */
  doUpload: function(aWindow) {
    const URL = "chrome://adblockfilters/content/ftp-upload.xul";

    var up = new UploadJob(this.getData());
    if(up.dataChanged) // we don't have to do anything
      aWindow.openDialog(URL, URL, "chrome", up, true);
  },
  isToolkitApp: function() {
    var appInfo = Cc["@mozilla.org/xre/app-info;1"]
                    .createInstance(Ci.nsIXULAppInfo);
    switch(appInfo.ID) {
      case "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}": // SeaMonkey
        var versionChecker = Cc["@mozilla.org/xpcom/version-comparator;1"]
                               .getService(Ci.nsIVersionComparator);
        return (versionChecker.compare("1.5a", appInfo.version) >= 0) &&
               ("nsIExtensionManager" in Ci) &&
               ("@mozilla.org/extensions/manager;1" in Cc);
      case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": // Firefox
      case "{a463f10c-3994-11da-9945-000d60ca027b}": // Flock
      case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}": // Netscape
      case "{3550f703-e582-4d05-9a08-453d09bdfdc6}": // Thunderbird
      case "{718e30fb-e89b-41dd-9da7-e25a45638b28}": // Sunbird
      case "{136c295a-4a5a-41cf-bf24-5cee526720d5}": // NVU
      case "songbird@songbirdnest.com":              // Songbird
      default:
        return true;
    }
  },
  openURL: function openURL(URL) {
    var appInfo = Cc["@mozilla.org/xre/app-info;1"]
                    .createInstance(Ci.nsIXULAppInfo);
    switch(appInfo.ID) {
      case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": // Firefox
      case "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}": // SeaMonkey
      case "{a463f10c-3994-11da-9945-000d60ca027b}": // Flock
      case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}": // Netscape
        var win = Cc["@mozilla.org/appshell/window-mediator;1"]
                    .getService(Ci.nsIWindowMediator)
                    .getMostRecentWindow("navigator:browser");

        const TAB_PREF = "browser.link.open_external";
        var tabsEnabled = nsPreferences.getIntPref(TAB_PREF) == 3;

        if(tabsEnabled && win) {
          win.focus();
          var browser = win.getBrowser();
          if(browser.selectedBrowser.currentURI.spec != "about:blank")
            browser.selectedTab = browser.addTab(URL);
          else
            browser.loadURI(makeURI(URL));
        } else {
          var windowFeatures = "chrome=0,resizable,modal=0,dialog=0,close,"
                              + "minimizable,titlebar,menubar,toolbar,location,"
                              + "directories,personalbar,status";
          win.open(URL, "_blank", windowFeatures);
        }
        break;
      case "{3550f703-e582-4d05-9a08-453d09bdfdc6}": // Thunderbird
      case "{718e30fb-e89b-41dd-9da7-e25a45638b28}": // Sunbird
      case "{136c295a-4a5a-41cf-bf24-5cee526720d5}": // NVU
      case "songbird@songbirdnest.com":              // Songbird
      default: // open in OS default
        Cc["@mozilla.org/uriloader/external-protocol-service;1"]
          .getService(Ci.nsIExternalProtocolService).loadUrl(makeURI(URL));
        break;
    }
  },
  get abpVersion() {
    if("@mozilla.org/adblockplus;1" in Cc)
      return Cc["@mozilla.org/adblockplus;1"]
               .createInstance().wrappedJSObject.getInstalledVersion();
    else
      return "0";
  },

  /* blacklist */
  _blacklist: null,
  get blacklist() {
    if(!this._blacklist) {
      var blacklist = [];
      var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService)
                    .getBranch("extensions.adblockfilters.blacklist.");
      var list = prefs.getChildList("", {});
      list.sort(function(a, b) { return a - b; });
      for(var i = 0; i < list.length; i++) {
        if(prefs.prefHasUserValue(list[i])) {
          blacklist.push(prefs.getComplexValue(list[i], Ci.nsISupportsString).data);
        }
      }
      this._blacklist = blacklist;
    }
    return this._blacklist;
  },
  set blacklist(blacklist) {
    var equal = this.blacklist.length == blacklist.length;
    for(var i = 0; i < this._blacklist.length && equal; i++) {
      equal = (this._blacklist[i] == blacklist[i]);
    }
    if(equal) return blacklist;

    var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService)
                  .getBranch("extensions.adblockfilters.blacklist.");
    var list = prefs.getChildList("", {});
    for(var i = 0; i < list.length; i++) {
      if(prefs.prefHasUserValue(list[i]) && !blacklist[list[i]])
        prefs.clearUserPref(list[i]);
    }

    var str = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
    for(var i = 0; i < blacklist.length; i++) {
      if(blacklist[i]) {
        str.data = blacklist[i];
        prefs.setComplexValue(i + 1, Ci.nsISupportsString, str);
      }
    };
    return this._blacklist = blacklist;
  },

  /* collect and assemble filters from abp */
  sortCallback: function(filter1, filter2) {
    if(filter1.type == "comment" && filter2.type == "comment")
      return 0;
    if(filter1.text < filter2.text)
      return -1;
    if (filter1.text > filter2.text)
      return 1;
    return 0;
  },
  getUserPatterns: function() {
    var abp = Cc["@mozilla.org/adblockplus;1"].createInstance().wrappedJSObject;
    var patterns = [];
    var doSort = nsPreferences.getBoolPref("extensions.adblockfilters.sortOutput", false);
    if(abp.prefs.userPatterns) {
      var prefs = abp.prefs;
      for(var i = 0; i < prefs.userPatterns.length; i++) {
        var pattern = {
          text: prefs.userPatterns[i].text,
          type: prefs.userPatterns[i].type,
          disabled: prefs.userPatterns[i].disabled,
          thirdParty: null, collapse: null,
          includeDomains: null, excludeDomains: null
        };
        patterns.push(pattern);
      }
      if(doSort)
        patterns = patterns.sort(this.sortCallback);
    } else {
      for each(var subscription in abp.filterStorage.knownSubscriptions) {
        if(subscription instanceof abp.SpecialSubscription) {
          var subPatterns = [];
          for each(var filter in subscription.filters) {
            var pattern = {
              text: filter.text,
              disabled: (filter.disabled || subscription.disabled),
              type: "unknown", thirdParty: filter.thirdParty, collapse: filter.collapse,
              includeDomains: filter.includeDomains, excludeDomains: filter.includeDomains
            };
            if(filter instanceof abp.CommentFilter) {
              pattern.type = "comment";
            } else if(filter instanceof abp.BlockingFilter || filter instanceof abp.RegExpFilter)
              pattern.type = "filterlist";
            else if(filter instanceof abp.WhitelistFilter)
              pattern.type = "whitelist";
            else if(filter instanceof abp.ElemHideFilter)
              pattern.type = "elemhide";
            subPatterns.push(pattern);
          }
          if(doSort)
            subPatterns = subPatterns.sort(this.sortCallback)
          patterns = patterns.concat(subPatterns);
        }
      }
    }
    return patterns;
  },
  getData: function(patterns, blacklist) {
    var abp = Cc["@mozilla.org/adblockplus;1"].createInstance().wrappedJSObject;
    function max(a, b) {
      return (abp.versionComparator.compare(a, b) > 0) ? a : b;
    }
    const MAX_VERSION = "1.0.1";
    function filterRequiresAbp(pattern) {
      var filter = pattern.text, type = filter.type;
      if(type == "comment" || type == "invalid")
        return "0";
      if(pattern.includeDomains != null || pattern.excludeDomains != null)
        return "1.0.1";
      if(pattern.thirdParty != null)
        return "1.0";
      if(pattern.collapse != null)
        return "0.7.5";
      if(type == "elemhide")
        return (/^#([\w\-]+|\*)(?:\(([\w\-]+)\))?$/.test(filter) ? "0.6.1" : "0.7");
      if((abp.Filter ? abp.Filter : abp).optionsRegExp.test(filter))
        return "0.7.1";
      if(/^(?:@@)\|?/.test(filter) || filter.substr(-1) == "|" || type == "whitelist")
        return "0.6.1.2";
      return "0";
    }

    var data = [];

    const COMMENT_PREFIX = "extensions.adblockfilters.comment.";
    var stringBundle = Cc["@mozilla.org/intl/stringbundle;1"]
                         .getService(Ci.nsIStringBundleService)
              .createBundle("chrome://adblockfilters/locale/upload.properties");
    if(nsPreferences.getBoolPref(COMMENT_PREFIX + "expire")) {
      var time = nsPreferences.getIntPref(COMMENT_PREFIX + "expire.value");
      data.push(stringBundle.formatStringFromName("comment-expire", [time], 1));
    }
    if(nsPreferences.getBoolPref(COMMENT_PREFIX + "redirect")) {
      var url = nsPreferences.copyUnicharPref(COMMENT_PREFIX + "redirect.value");
      data.push(stringBundle.formatStringFromName("comment-redirect", [url], 1));
    }

    if(!patterns) patterns = this.getUserPatterns();
    if(!blacklist) blacklist = this.blacklist;
    var text, minVersion = "0";
    for(var i = 0; i < patterns.length; i++) {
      text = patterns[i].text;
      if(blacklist.indexOf(text) != -1)
        continue;
      if(patterns[i].disabled && patterns[i].type != "comment") {
        text = "!" + text;
      } else if(minVersion != MAX_VERSION) {
        minVersion = max(filterRequiresAbp(patterns[i]), minVersion);
      }
      data.push(text);
    }


    if(max(minVersion, "0.7.1pre") != "0.7.1pre")
      data.splice(0, 0, "[Adblock Plus " + minVersion + "]");
    else if(minVersion != "0")
      data.splice(0, 0, "(Adblock Plus " + minVersion + " or higher required) " + data[0]);
    else
      data.splice(0, 0, "[Adblock]");

    return data;
  },
  serializeArray: function(data, replaceVars) {
    if(replaceVars)
      data = this.replaceVariables(data);

    var abp = Cc["@mozilla.org/adblockplus;1"].createInstance().wrappedJSObject;
    return data.join(abp.getLineBreak());
  },
  getHash: function(data) {
    // normalize the input data
    data = data.replace(/\r+/g, "").replace(/\n+/, "\n");

    var converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                      .createInstance(Ci.nsIScriptableUnicodeConverter);
    converter.charset = "UTF-8";
    data = converter.convertToByteArray(data, {});

    var ch = Cc["@mozilla.org/security/hash;1"].createInstance(Ci.nsICryptoHash);
    ch.init(ch.MD5);
    ch.update(data, data.length);
    return ch.finish(true).replace(/=+$/, "");
  },
  dataChanged: function(hash, data) {
    const PREF = "extensions.adblockfilters.lastUploadHash";
    if(!data) data = this.getData();
    if(!hash) hash = this.getHash(this.serializeArray(data, false));
    return (nsPreferences.copyUnicharPref(PREF) != hash);
  },
  replaceVariables: function(data) {
    var now = "$1" + new Date().toLocaleString() + "$2";
    return data.map(function(pattern) {
      if(pattern[0] != "!")
        return pattern;

      return pattern.replace(/(\s)\$DATE(\s|$)/g, now);
    });
  },


  // nsISupports
  QueryInterface: function(aIID) {
    if(aIID.equals(Ci.nsISupports) || aIID.equals(Ci.nsIClassInfo) ||
       aIID.equals(Ci.nsIObserver))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE;
  },

  // nsIClassInfo
  getHelperForLanguage: function() {
    return null;
  },
  getInterfaces: function getInterfaces(aCount) {
    var array = [Ci.nsIClassInfo, Ci.nsIObserver];
    aCount.value = array.length;
    return array;
  },
  //classDescription: "some helpers for the Adblock Plus Filter Uploader extension... nothing you could need",
  classIDNoAlloc: Ci.nsISupports,
  flags: Ci.nsIClassInfo.SINGLETON,
  implementationLanguage: Ci.nsIProgrammingLanguage.JAVASCRIPT
};


/*
 * Provides data and target URI to upload to.
 */
function UploadJob() {
  this.wrappedJSObject = this;
  if(arguments.length)
    this.init.apply(this, arguments);
}
UploadJob.prototype = {
  classID: Components.ID("{3e0cb45d-4e51-fdf2-839b-5158df0c4d15}"),
  contractID: "@adblockfilters.mozdev.org/uploadjob;1",
  classDescription: "Adblock Plus Filter Uploader upload job",

  wrappedJSObject: null,

  dataChanged: true,
  init: function(data, window, onNotUploaded) {
    var service = Components.classes["@adblockfilters.mozdev.org/service;1"]
                            .getService().wrappedJSObject;

    this.writeHash = service.getHash(service.serializeArray(data, false));
    data = service.serializeArray(data, true);

    // add a line with the checksum:
    // xxx this shouldn't be in the upload code
    data = data.replace(/([\n\r]+)/, "$1! Checksum: " + service.getHash(data) + "$1");

    this.data = data;

    this.dataChanged = service.dataChanged(this.writeHash);

    if(window)        this.origWindow = window;
    if(onNotUploaded) this.onNotUploaded = onNotUploaded;

    this.compress();
  },
  windowDestroyed: function() {
    // fall back on the prototype's defaults
    delete this.onNotUploaded;
    delete this.origWindow;
  },

  get targetURI() {
    const PREF_IX = "extensions.adblockfilters.ftp.";
    var protocol = nsPreferences.copyUnicharPref(PREF_IX + "protocol");
    var host = nsPreferences.copyUnicharPref(PREF_IX + "host");
    var path = nsPreferences.copyUnicharPref(PREF_IX + "path");

    if(!host || !path) throw Cr.NS_ERROR_NOT_AVAILABLE;

    if(path[0] != "/") path = "/" + path; // xxx?

    return makeURI(protocol + "://" + host + path);
  },
  getUsernameAndPassword: function() {
    const PREF_IX = "extensions.adblockfilters.ftp.";
    var username = nsPreferences.copyUnicharPref(PREF_IX + "user");
    var password = null;
    try {
      if("nsILoginManager" in Ci && username) {
        var passwordManager = Cc["@mozilla.org/login-manager;1"]
                                .getService(Ci.nsILoginManager);
        // Find users for the given parameters
        var pseudoURI = "chrome://adblockfilters";
        var logins = passwordManager.findLogins({}, pseudoURI, pseudoURI, null);
        // Find user from returned array of nsILoginInfo objects
        for(var i = 0; i < logins.length; i++) {
          if(logins[i].username == username) {
            password = logins[i].password;
            break;
          }
        }
      } else if(username) {
        // This doesn't harm if passwords are not encrypted
        var tokendb = Cc["@mozilla.org/security/pk11tokendb;1"]
                        .createInstance(Ci.nsIPK11TokenDB);
        var token = tokendb.getInternalKeyToken();

        if(!token.checkPassword("")) {
          // So there's a master password. But since checkPassword didn't
          // succeed, we're logged out (per nsIPK11Token.idl).
          // Relogin and ask for the master password.
          token.login(true);  // 'true' means always prompt for token password.
                              // User will be prompted until clicking 'Cancel'
                              // or entering the correct password.

          // An exception will be thrown if the user cancels the login prompt
          // dialog. User is also logged out of Software Security Device.
        }

        var passwordManager = Cc["@mozilla.org/passwordmanager;1"]
                                .createInstance(Ci.nsIPasswordManager);
        var passwords = passwordManager.enumerator;
        while(passwords.hasMoreElements()) {
          var nsIPassword = passwords.getNext().QueryInterface(Ci.nsIPassword);
          if(nsIPassword.host == host && nsIPassword.user == username) {
            password = nsIPassword.password;
            break;
          }
        }
      }
    } catch(e) { Components.reportError(e); }
    if(!username || password === null) {
      var prompts = Cc["@mozilla.org/embedcomp/prompt-service;1"]
                      .getService(Ci.nsIPromptService);

      var sbs = Cc["@mozilla.org/intl/stringbundle;1"]
                  .getService(Ci.nsIStringBundleService);
      var stringBundle = sbs.createBundle("chrome://adblockfilters/locale/upload.properties");
      var title = stringBundle.GetStringFromName("window-title");
      var text = stringBundle.GetStringFromName("server-auth");

      username = { value: username };
      password = { value: password || "" };
      if(!prompts.promptUsernameAndPassword(null, title, text, username,
                                            password, null, {}))
        return null;
      username = username.value;
      password = password.value;
    }
    return {user: username, password: password};
  },

  _compressedData: "",
  get streamToUpload() {
    var storageStream = Cc["@mozilla.org/storagestream;1"]
                          .createInstance(Ci.nsIStorageStream);
    storageStream.init(8192, -1, null);
    var outputStream = storageStream.getOutputStream(0);
    if(this._compressedData) {
      outputStream.write(this._compressedData, this._compressedData.length);
      outputStream.close();
    } else {
      var os = Cc["@mozilla.org/intl/converter-output-stream;1"]
                 .createInstance(Ci.nsIConverterOutputStream);
      os.init(outputStream, "UTF-8", 8192, "?".charCodeAt(0));
      os.writeString(this.data);
      os.close();
    }
    return storageStream.newInputStream(0);
  },

  compress: function() {
    if(!("@mozilla.org/streamconv;1?from=uncompressed&to=rawdeflate" in Cc)) return;
    const PREFIX = "extensions.adblockfilters.comp.";
    if(!nsPreferences.getBoolPref(PREFIX + "enabled")) return;

    var streamListener = {
      onStartRequest: function(request, context) { },
      onDataAvailable: function(request, context, stream, offset, count) {
        // stream seems to implement nsISupportsString even though instanceof says differently:
        context.wrappedJSObject.data += stream.data.substr(0, count);
      },
      onStopRequest: function(request, context, statusCode) { }
    };
    var streamConverter = Cc["@mozilla.org/streamConverters;1"]
                            .getService(Ci.nsIStreamConverterService);
    var deflate = { data : "" }; deflate.wrappedJSObject = deflate;
    var compressor = streamConverter.asyncConvertData("uncompressed", "rawdeflate",
                                                      streamListener, deflate);
    var stream = this.streamToUpload;
    compressor.onStartRequest(null, null);
    compressor.onDataAvailable(null, null, stream, 0, stream.available());
    compressor.onStopRequest(null, null, Cr.NS_OK);
    stream.close();

    // convert the js string to UTF-8, so we can compute its hash
    var bis = Cc["@mozilla.org/binaryinputstream;1"]
                .createInstance(Ci.nsIBinaryInputStream);
    bis.setInputStream(this.streamToUpload);
    var data = bis.readBytes(bis.available());
    bis.close();
    // gzip is deflate with some additional header data, see rfc 1952
    loadScriptOnce("crc32.js");
    const GZIP_HEADER = "\x1f\x8b\x08\x01\0\0\0\0\0\0";
    var crc = numToByteStr(crc32(data), 4);
    var size = numToByteStr(data.length, 4);
    var gzip = GZIP_HEADER + deflate.data + crc + size;
    this._compressedData = gzip;
    this.contentType = "application/gzip";
  },

  contentType: "text/plain",

  data: null,
  writeHash: null,
  origWindow: null,
  onUploaded: function onUploaded() {
    const PREF = "extensions.adblockfilters.lastUploadHash";
    nsPreferences.setUnicharPref(PREF, this.writeHash);

    var origWindow = this.origWindow;
    if(origWindow && !origWindow.closed)
      origWindow.document.documentElement.getButton("extra1").disabled = true;
  },
  onNotUploaded: function() {},

  // nsISupports
  QueryInterface: function(aIID) {
    if(aIID.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE;
  }
};


// constructors for objects we want to XPCOMify
var objects = [AdblockFilterService, UploadJob];


/*
 * Common registration code.
 *
 */


function FactoryHolder(aObj) {
  this.CID        = aObj.prototype.classID;
  this.contractID = aObj.prototype.contractID;
  this.className  = aObj.prototype.classDescription;
  this.factory = {
    createInstance: function(aOuter, aIID) {
      if(aOuter)
        throw Cr.NS_ERROR_NO_AGGREGATION;

      return (new aObj()).QueryInterface(aIID);
    }
  };
}

var gModule = {
  registerSelf: function (aComponentManager, aFileSpec, aLocation, aType)
  {
    aComponentManager.QueryInterface(Ci.nsIComponentRegistrar);
    for (var key in this._objects) {
      var obj = this._objects[key];
      aComponentManager.registerFactoryLocation(obj.CID, obj.className,
                                   obj.contractID, aFileSpec, aLocation, aType);
    }
  },

  unregisterSelf: function(aCompMgr, aFileSpec, aLocation) {
    aComponentManager.QueryInterface(Ci.nsIComponentRegistrar);
    for (var key in this._objects) {
      var obj = this._objects[key];
      aComponentManager.unregisterFactoryLocation(obj.CID, aFileSpec);
    }
  },

  getClassObject: function(aComponentManager, aCID, aIID) {
    if (!aIID.equals(Ci.nsIFactory)) throw CR.NS_ERROR_NOT_IMPLEMENTED;

    var key = aCID.number;
    if (key in this._objects && aCID.equals(this._objects[key].CID))
      return this._objects[key].factory;

    throw CR.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aComponentManager) {
    return true;
  },

  _objects: {} //FactoryHolder
};

function NSGetModule(compMgr, fileSpec)
{
  for(var i = 0; i < objects.length; i++) {
    var obj = new FactoryHolder(objects[i]);
    gModule._objects[obj.CID.number] = obj;
  }
  return gModule;
}